from typing import Dict, Optional, Union

import pytest

from pipeline_lib.pipeline_task import PipelineTask
from pipeline_lib.type_checking import (
    PipelineTypeError,
    get_func_args,
    is_iterable,
    type_check_tasks,
)

from .example_funcs import *


def test_get_func_args():
    assert get_func_args(generate_numbers) == (None, int, [])
    assert get_func_args(group_numbers) == (int, List[int], ["num_groups"])
    assert get_func_args(print_numbers) == (int, None, [])

    # test kw only arguments
    def kwarg_func(x: Iterable[int], arg2: float, *, arg3: str) -> Iterable[str]:
        pass

    assert get_func_args(kwarg_func) == (int, str, ["arg2", "arg3"])
    assert get_func_args(kwarg_func, extract_first=False) == (
        None,
        str,
        ["x", "arg2", "arg3"],
    )

    # errors
    with pytest.raises(PipelineTypeError):

        def no_return_type():
            pass

        get_func_args(no_return_type)

    with pytest.raises(PipelineTypeError):

        def first_argument_not_iterable(x: Union[str, float]) -> Iterable[str]:
            pass

        get_func_args(first_argument_not_iterable)

    with pytest.raises(PipelineTypeError):

        def return_not_iterable(x: Iterable[str]) -> Optional[str]:
            pass

        get_func_args(return_not_iterable)


def test_is_iterable():
    assert is_iterable(Iterable[str])
    assert not is_iterable(Optional[str])
    assert not is_iterable(float)


def test_type_checks_valid():
    tasks = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(group_numbers, constants={"num_groups": 5}),
        PipelineTask(
            sum_numbers,
        ),
        PipelineTask(
            print_numbers,
        ),
    ]
    type_check_tasks(tasks)


def test_mismatched_task_types():
    mismatched_tasks = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            sum_numbers,
        ),
        PipelineTask(
            print_numbers,
        ),
    ]
    with pytest.raises(PipelineTypeError):
        type_check_tasks(mismatched_tasks)


def test_needed_none_start():
    needed_none_start = [
        PipelineTask(
            print_numbers,
        )
    ]
    with pytest.raises(PipelineTypeError):
        type_check_tasks(needed_none_start)


def start_with_consts(beginning: Dict[int, str]) -> Iterable[int]:
    pass


def sink_ints(vals: Iterable[int]) -> None:
    list(vals)


def test_start_with_consts():
    # check that first function can accept constant arguments in its first argument
    start_with_consts_t = [
        PipelineTask(start_with_consts, constants={"beginning": {1: "bob"}}),
        PipelineTask(sink_ints),
    ]
    type_check_tasks(start_with_consts_t)


def test_non_in_middle():
    none_in_middle = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            print_numbers,
        ),
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            print_numbers,
        ),
    ]
    with pytest.raises(PipelineTypeError):
        type_check_tasks(none_in_middle)


def test_last_is_not_none_task_check():
    last_is_not_none_tasks = [
        PipelineTask(
            generate_numbers,
        )
    ]
    with pytest.raises(PipelineTypeError):
        type_check_tasks(last_is_not_none_tasks)


def test_last_is_none_task_check():
    last_is_none_tasks = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            print_numbers,
        ),
    ]
    type_check_tasks(last_is_none_tasks)


def test_consts_present_check():
    consts_present = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            add_const,
            constants={
                "add_val": 5,
            },
        ),
        PipelineTask(
            print_numbers,
        ),
    ]
    type_check_tasks(consts_present)


def test_consts_missing_check():
    consts_missing = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(add_const, constants={}),
        PipelineTask(
            print_numbers,
        ),
    ]
    with pytest.raises(PipelineTypeError):
        type_check_tasks(consts_missing)


def test_consts_added_check():
    consts_added = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            add_const,
            constants={
                "add_val": 5,
                "extra_val": 6,
            },
        ),
        PipelineTask(
            print_numbers,
        ),
    ]
    with pytest.raises(PipelineTypeError):
        type_check_tasks(consts_added)



import warnings
from typing import List

from .pipeline_task import (
    DEFAULT_BUF_SIZE,
    DEFAULT_NUM_WORKERS,
    DEFAULT_PACKETS_IN_FLIGHT,
    PipelineTask,
)


def _warn_parameter_overrides(tasks: List[PipelineTask]):
    for task in tasks:
        if (
            task.max_message_size != DEFAULT_BUF_SIZE
            or task.num_workers != DEFAULT_NUM_WORKERS
            or task.packets_in_flight != DEFAULT_PACKETS_IN_FLIGHT
        ):
            warnings.warn(
                f"Task '{task.name}' overrode default value of max_message_size, num_workers, or packets_in_flight, and this override is ignored by 'coroutine' parallelism strategy."
            )


def execute_seq(tasks: List[PipelineTask]):
    # pylint: disable=too-many-branches,too-many-locals
    """
    execute tasks until final task completes.
    Raises error if tasks are inconsistently specified or if
    one of the tasks raises an error.
    """
    if not tasks:
        return

    _warn_parameter_overrides(tasks)

    cur_result = tasks[0].generator(**tasks[0].constants_dict)
    for task in tasks[1:]:
        cur_result = task.generator(cur_result, **task.constants_dict)



import inspect
import typing
from typing import Iterable, List, Type

from .pipeline_task import PipelineTask

MAX_NUM_WORKERS = 128


class PipelineTypeError(RuntimeError):
    pass


def _type_error_if(condition, message):
    if not condition:
        raise PipelineTypeError(message)


def is_iterable(type_: Type):
    return typing.get_origin(type_) is typing.get_origin(Iterable)


def get_func_args(func, extract_first=True):
    arguments = inspect.getfullargspec(func)
    _type_error_if(arguments.varargs is None, "varargs not supported")
    _type_error_if(arguments.varkw is None, "varkw not supported")
    _type_error_if(arguments.defaults is None, "default arguments not supported")
    _type_error_if(arguments.kwonlydefaults is None, "default arguments not supported")
    _type_error_if(
        set(arguments.args + arguments.kwonlyargs).issubset(arguments.annotations),
        "all arguments must have annotations",
    )
    _type_error_if(
        "return" in arguments.annotations,
        "function return type must have type annotation, if not returning, please specify ->None",
    )

    base_input_type = (
        None
        if not arguments.args or not extract_first
        else arguments.annotations[arguments.args[0]]
    )
    base_return_type = arguments.annotations["return"]

    _type_error_if(
        base_input_type is None
        or (
            is_iterable(base_input_type) and len(typing.get_args(base_input_type)) == 1
        ),
        "First argument must be an Iterable[input_type], if defined",
    )
    _type_error_if(
        base_return_type is None
        or (
            is_iterable(base_return_type)
            and len(typing.get_args(base_return_type)) == 1
        ),
        f"Return type annotation must be an Iterable[input_type] or None, was {base_return_type}",
    )

    input_type = (
        None if base_input_type is None else typing.get_args(base_input_type)[0]
    )
    return_type = (
        None if base_return_type is None else typing.get_args(base_return_type)[0]
    )

    # these are guarentteed to be mutually exclusive
    other_argument_names = arguments.args + arguments.kwonlyargs
    # remove input argument
    if arguments.args and extract_first:
        other_argument_names.remove(arguments.args[0])

    return input_type, return_type, other_argument_names


def type_check_tasks(tasks: List[PipelineTask]):
    prev_type = None
    for task_idx, task in enumerate(tasks):
        input_type, return_type, other_args = get_func_args(
            task.generator, extract_first=(task_idx != 0)
        )
        if prev_type != input_type:
            raise PipelineTypeError(
                f"In task {task.name}, expected input {input_type}, received input {prev_type}."
            )

        if task_idx != len(tasks) - 1 and return_type is None:
            raise PipelineTypeError(
                "None return type not allowed in any task except final task of pipe"
            )

        task_consts = {} if task.constants is None else task.constants
        task_const_names = list(task_consts.keys())
        if set(task_consts) != set(other_args):
            raise PipelineTypeError(
                f"In task {task.name}, expected constants {other_args}, received constants {task_const_names}."
            )

        prev_type = return_type

    if prev_type is not None:
        raise PipelineTypeError(
            f"In final task {tasks[-1].name}, expected output type None, actual type {prev_type}."
        )


def sanity_check_mp_params(tasks: List[PipelineTask]):
    for task in tasks:
        _sanity_check_mp_params(task)


def _sanity_check_mp_params(task: PipelineTask):
    if task.num_workers <= 0:
        raise PipelineTypeError(
            f"In task {task.name}, num_workers value {task.num_workers} needs to be positive"
        )

    if task.num_workers > MAX_NUM_WORKERS:
        raise PipelineTypeError(
            f"In task {task.name}, num_workers value {task.num_workers} was greater than hard limit {MAX_NUM_WORKERS} for number of threads per task"
        )

    if task.num_workers > task.packets_in_flight:
        raise PipelineTypeError(
            f"In task {task.name}, packets_in_flight {task.packets_in_flight} is less than num_workers {task.num_workers}, which can lead to deadlocks."
        )

    if task.shared_buffer and task.max_message_size is None:
        raise PipelineTypeError(
            f"In task {task.name}, shared_buffer={task.shared_buffer}, but max_message_size is None, which has no effect, as shared_buffer only operates when the max_message_size is set"
        )



import time
import typing
from contextlib import contextmanager
from typing import Iterable, List

import pytest

from pipeline_lib.execution import ParallelismStrategy

all_parallelism_options: List[ParallelismStrategy] = typing.get_args(
    ParallelismStrategy
)
thread_parallelism_options: List[ParallelismStrategy] = [
    "thread",
    "process-fork",
    "process-spawn",
]
process_parallelism_options: List[ParallelismStrategy] = [
    "process-fork",
    "process-spawn",
]


def sleeper(vals: Iterable[int], sleep_time: float) -> Iterable[int]:
    time.sleep(0.1)
    for i in vals:
        time.sleep(sleep_time)
        yield i



from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional

DEFAULT_BUF_SIZE = 131072
DEFAULT_NUM_WORKERS = 1
DEFAULT_PACKETS_IN_FLIGHT = 1


@dataclass
class PipelineTask:
    """
    Definition of a task to place in the pipeline
    """

    generator: Callable
    constants: Optional[Dict[str, Any]] = None
    num_workers: int = DEFAULT_NUM_WORKERS
    # one packet in flight means that between both the producer and consumer,
    # only one packet ever is being processed at a time.
    # So one packet means full execution synchronization
    packets_in_flight: int = DEFAULT_PACKETS_IN_FLIGHT
    # only applicable in multiprocessing setting
    # if set to None, then no limit on message size, but
    # message reads will be much slower
    max_message_size: Optional[int] = None
    # when max_message_size is set, the downstream pipeline reads memory from
    # a shared buffer without copying. This is valid if the downstream pipeline step
    # deletes all references to all the data it receives from the previous iteration
    # of the iterable
    # default is to copy this buffer to guarantee no data races
    shared_buffer: bool = False

    @property
    def name(self) -> str:
        return self.generator.__name__

    @property
    def constants_dict(self) -> Dict[str, Any]:
        return {} if self.constants is None else self.constants


class TaskError(RuntimeError):
    """Error for miscellaneous, unidentifiable issues that come up during task execution"""


class InactivityError(RuntimeError):
    """Error for when the inactivity timeout expires without any messages being passed"""



import multiprocessing as mp
import os
import signal
import time
from multiprocessing import synchronize
from typing import Any, Dict

import numpy as np
import psutil
import pytest
from pytest import raises

import pipeline_lib
from pipeline_lib import PipelineTask, execute
from pipeline_lib.execution import ParallelismStrategy
from pipeline_lib.pipeline_task import InactivityError

from .example_funcs import *
from .test_utils import (
    all_parallelism_options,
    process_parallelism_options,
    sleeper,
    thread_parallelism_options,
)


class TestExpectedException(ValueError):
    pass


def raise_exception_fn(arg: Iterable[int]) -> Iterable[int]:
    # start up input generator/process
    i1 = next(iter(arg))
    yield i1
    raise TestExpectedException()


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_execute_exception(parallelism: ParallelismStrategy):
    tasks = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            raise_exception_fn,
        ),
        PipelineTask(
            print_numbers,
        ),
    ]
    with raises(TestExpectedException):
        execute(tasks, parallelism)


class SuddenExit(RuntimeError):
    pass


def sudden_exit_fn(arg: Iterable[int]) -> Iterable[int]:
    # start up input generator/process
    next(iter(arg))
    # thread raises exception so that python does not know about it
    raise SuddenExit("sudden exit")


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_sudden_exit_middle(parallelism: ParallelismStrategy):
    tasks = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            sudden_exit_fn,
        ),
        PipelineTask(
            print_numbers,
        ),
    ]
    with raises(SuddenExit):
        execute(tasks, parallelism)


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_sudden_exit_end(parallelism: ParallelismStrategy):
    tasks = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            sudden_exit_fn,
        ),
        PipelineTask(print_numbers),
    ]
    with raises(SuddenExit):
        execute(tasks, parallelism)


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_sudden_exit_middle_sleepers(parallelism: ParallelismStrategy):
    tasks = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(sleeper, constants={"sleep_time": 0.1}),
        PipelineTask(
            sudden_exit_fn,
        ),
        PipelineTask(sleeper, constants={"sleep_time": 0.1}),
        PipelineTask(
            print_numbers,
        ),
    ]
    with raises(SuddenExit):
        execute(tasks, parallelism)


def generate_numbers_short() -> Iterable[int]:
    for i in range(9):
        yield i


@pytest.mark.parametrize("parallelism", thread_parallelism_options)
def test_inactivty_timeout(parallelism: ParallelismStrategy):
    """
    If we sleep for 1 second and have a task timeout of 0.1 seconds,
    we should error due to the task timeout
    """
    tasks = [
        PipelineTask(
            generate_numbers_short,
        ),
        PipelineTask(sleeper, constants={"sleep_time": 1}),
        PipelineTask(
            print_numbers,
        ),
    ]
    with raises(InactivityError):
        execute(tasks, parallelism, inactivity_timeout=0.1)


@pytest.mark.parametrize("parallelism", thread_parallelism_options)
def test_inactivity_timeout_missed(parallelism: ParallelismStrategy):
    """
    If we sleep for 0.1 second and have a task timeout of 1 seconds,
    we should not error due to the task timeout
    """
    tasks = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(sleeper, constants={"sleep_time": 0.1}),
        PipelineTask(
            print_numbers,
        ),
    ]
    # pipeline step should take about 10 seconds, 100 iters of 0.1 seconds each, so
    # this catches that it is only inactivity
    execute(tasks, parallelism, inactivity_timeout=5)


def consume_infinite_ints(vals: Iterable[int]) -> None:
    for _ in vals:
        pass


def start_pipeline(parallelism: ParallelismStrategy):
    execute(
        [
            PipelineTask(
                generate_infinite,
            ),
            PipelineTask(consume_infinite_ints, num_workers=2, packets_in_flight=2),
        ],
        parallelism=parallelism,
    )


@pytest.mark.parametrize("parallelism", all_parallelism_options)
@pytest.mark.parametrize(
    "sig,returnval",
    [
        (signal.SIGTERM, -15),
        (signal.SIGINT, 1),
    ],
)
def test_main_process_signal(
    sig: signal.Signals, returnval: int, parallelism: ParallelismStrategy
):
    """
    If main process receives a sigterm signal,
    all the other processes should exit cleanly and quickly
    """
    ctx = mp.get_context("spawn")
    proc = ctx.Process(target=start_pipeline, args=(parallelism,))
    proc.start()
    # wait for worker processes to start up
    time.sleep(10.0)

    # collect all living child processes
    child_procs = psutil.Process(proc.pid).children()

    process_parallelism = set(process_parallelism_options)
    if parallelism in process_parallelism:
        # there are 3 child processes we expect
        # 1 generate_infinite process, 2 consume_infinite_ints processes
        assert len(child_procs) == 3
    else:
        # in thread or coroutine, there won't be any child processes, that is expected
        assert len(child_procs) == 0

    # send a sigterm signal to the main process
    os.kill(proc.pid, sig)

    # waits for all the processes to shut down
    proc.join(20.0)
    assert (
        proc.exitcode is not None
    ), "join timed out, main process did not exist promptly after signterm"
    assert (
        proc.exitcode == returnval
    ), f"main process should return a {returnval} error code, returned {proc.exitcode}"

    for proc in child_procs:
        assert not psutil.pid_exists(
            proc.pid
        ), "main process didn't exit and join children during its shutdown process"


def only_error_if_second_proc(
    arg: Iterable[int], started_event: synchronize.Event
) -> Iterable[int]:
    """
    only exits if it is the first worker process to start up.
    """
    yield next(iter(arg))
    is_second_proc = started_event.is_set()
    started_event.set()
    if is_second_proc:
        raise TestExpectedException()
    else:
        yield from arg


def generate_infinite() -> Iterable[int]:
    yield from range(10000000000000)


@pytest.mark.parametrize("parallelism", thread_parallelism_options)
def test_single_worker_error(parallelism: ParallelismStrategy):
    """
    if one process dies and the others do not, then it should still raise an exception,
    as the dead process might have consumed an important message
    """
    mp_context = mp.get_context("spawn") if parallelism == "process-spawn" else mp
    started_event = mp_context.Event()
    tasks = [
        PipelineTask(
            generate_infinite,
        ),
        PipelineTask(
            only_error_if_second_proc,
            constants={
                "started_event": started_event,
            },
            num_workers=2,
            packets_in_flight=10,
        ),
        PipelineTask(print_numbers, num_workers=2, packets_in_flight=2),
    ]
    with raises(TestExpectedException):
        execute(tasks, parallelism)


def force_exit_if_second_proc(
    arg: Iterable[int], started_event: synchronize.Event
) -> Iterable[int]:
    """
    only exits if it is the first worker process to start up.
    """
    yield next(iter(arg))
    is_second_proc = started_event.is_set()
    started_event.set()
    if is_second_proc:
        # kill process using very low level os utilities
        # so that python does not know anything about process exiting
        os.kill(os.getpid(), signal.SIGKILL)
    else:
        yield from arg


@pytest.mark.parametrize("parallelism", process_parallelism_options)
def test_single_worker_unexpected_exit(parallelism: ParallelismStrategy):
    """
    if one process dies and the others do not, then it should still raise an exception,
    as the dead process might have consumed an important message
    """
    process_context_map: Dict[ParallelismStrategy, str] = {
        "process-fork": "fork",
        "process-spawn": "spawn",
    }
    ctx = mp.get_context(process_context_map[parallelism])
    started_event = ctx.Event()
    tasks = [
        PipelineTask(
            generate_infinite,
        ),
        PipelineTask(
            force_exit_if_second_proc,
            constants={
                "started_event": started_event,
            },
            num_workers=2,
            packets_in_flight=10,
        ),
        PipelineTask(print_numbers, num_workers=2, packets_in_flight=2),
    ]
    with raises(pipeline_lib.pipeline_task.TaskError):
        execute(tasks, parallelism)


def hang_message_passing() -> Iterable[int]:
    for i in range(8):
        yield i
    exit(0)


# if it takes more than 120 seconds for a 5 second timeout to complete, something is wrong
@pytest.mark.timeout(120)
@pytest.mark.parametrize("parallelism", process_parallelism_options)
@pytest.mark.parametrize("max_message_size", [10000, None])
def test_hang_message_passing_timeout(
    max_message_size: bool,
    parallelism: ParallelismStrategy,
):
    n_procs = 2
    packets_in_flight = 4
    tasks = [
        PipelineTask(
            hang_message_passing,
            max_message_size=max_message_size,
            packets_in_flight=packets_in_flight,
        ),
        PipelineTask(
            group_numbers,
            constants={"num_groups": 3},
            max_message_size=max_message_size,
            packets_in_flight=packets_in_flight,
            num_workers=n_procs,
        ),
        PipelineTask(
            sum_numbers,
            max_message_size=max_message_size,
            packets_in_flight=packets_in_flight,
            num_workers=n_procs,
        ),
        PipelineTask(print_numbers),
    ]
    with raises(InactivityError):
        execute(tasks, parallelism, inactivity_timeout=5)


if __name__ == "__main__":
    test_hang_message_passing_timeout(1000, "process-fork")



import contextlib
import ctypes
import logging
import multiprocessing as mp
import multiprocessing.connection as mp_connection
import os
import queue
import select
import signal
import sys
import threading as tr
import time
import traceback
import typing
from dataclasses import dataclass
from functools import reduce
from multiprocessing import synchronize
from multiprocessing.context import (
    BaseContext,
    ForkContext,
    ForkProcess,
    SpawnContext,
    SpawnProcess,
)
from operator import mul
from typing import Any, Iterable, List, Literal, Optional, Set, Tuple, Union

import cloudpickle

from .pipeline_task import InactivityError, PipelineTask, TaskError
from .type_checking import MAX_NUM_WORKERS, sanity_check_mp_params

logger = logging.getLogger(__name__)

ERR_BUF_SIZE = 2**16
# some arbitrary, hopefully unused number that signals python exiting after placing the error in the queue
PYTHON_ERR_EXIT_CODE = 187
# copies are much faster if they are aligned to 16 byte or 32 byte boundaries (depending on architecture)
ALIGN_SIZE = 32

SpawnContextName = Literal["spawn", "fork", "forkserver"]


class PropogateErr(RuntimeError):
    # should never be raised in main scope, meant to act as a proxy
    # for propogating errors up and down the pipeline
    pass


class SignalReceived(Exception):
    def __init__(self, signum: int) -> None:
        self.signum = signum


@dataclass
class Value:
    value: Any


def roundup_to_align(size):
    return size + (-size) % ALIGN_SIZE


class CyclicAllocator:
    # pylint: disable=too-many-instance-attributes
    def __init__(self, max_num_elements: int, ctx: BaseContext) -> None:
        self.max_num_elements = max_num_elements
        self.producer_idxs = ctx.RawArray(ctypes.c_int, self.max_num_elements)
        self.producer_next = ctx.Value(ctypes.c_int, 0, lock=False)
        self.producer_last = ctx.Value(ctypes.c_int, 0, lock=False)
        # initialize to putting everything available
        for i in range(self.max_num_elements):
            self.producer_idxs[i] = i

        self.consumer_idxs = ctx.RawArray(ctypes.c_int, self.max_num_elements)
        self.consumer_next = ctx.Value(ctypes.c_int, 0, lock=False)
        self.consumer_last = ctx.Value(ctypes.c_int, 0, lock=False)

        self.position_lock = ctx.Lock()

    def pop_write_pos(self):
        with self.position_lock:
            write_entry = int(self.producer_next.value)
            write_pos = int(self.producer_idxs[write_entry])
            self.producer_next.value = (write_entry + 1) % self.max_num_elements
        return write_pos

    def push_read_pos(self, written_pos: int):
        with self.position_lock:
            write_entry = int(self.consumer_last.value)
            self.consumer_idxs[write_entry] = written_pos
            self.consumer_last.value = (write_entry + 1) % self.max_num_elements  # type: ignore

    def pop_read_pos(self):
        with self.position_lock:
            read_entry = int(self.consumer_next.value)
            last_read_entry = int(self.consumer_last.value)
            # print(read_entry, last_read_entry)
            if read_entry == last_read_entry:
                # no entry available for reading, cannot continue
                raise queue.Empty()
            read_pos = self.consumer_idxs[read_entry]
            self.consumer_next.value = (read_entry + 1) % self.max_num_elements
        return read_pos

    def push_write_pos(self, read_pos):
        with self.position_lock:
            read_entry = int(self.producer_last.value)
            self.producer_idxs[read_entry] = read_pos
            self.producer_last.value = (read_entry + 1) % self.max_num_elements


class AsyncQueue:
    def get(self) -> Tuple[Any, int]:
        """
        returns a tuple of (queued_item, packet_id)

        Need to call free(packet_id) after finished using the object.

        If no objects to get, raises queue.Empty error
        """
        raise NotImplementedError()

    def put(self, _item: Any):
        """
        Puts item on queue. Undefined behavior if there are more items put
        on the queue than the space available.

        TODO: raise an error
        """
        raise NotImplementedError()

    def free(self, _packet_id: int):
        """
        Free up fetched packet for writing
        """
        raise NotImplementedError()

    def flush(self, has_error: synchronize.Event):
        """
        Flush all pending writes (will block unless has_error is set)
        """


END_OF_BUFFER_ID = -2


class BufferedQueue(AsyncQueue):
    """
    A custom queue implementation based on fixed-size shared memory buffers to transfer data.

    Advantages over standard library multiprocessing.Queue:

    1. Objects, once placed on the queue can be fetched immediately without delay (mp.Queue `put` function returns immediately after spawning the "sender" thread, so there can be a delay before the message is readable from the queue)
    2. Even if producer process dies, the message is still available to consume (only possible in ordinary queue if the message is placed fully in the mp.Pipe hardcoded 64kb buffer)
    3. No extra thread (on producer) is needed to send data to consumer
    4. Buffered communication between processes occurs fully asynchronously (vs back and forth queue message passing between spawned producer thread and consumer)
    5. One queue buffered read can process concurrently with one write

    Disadvantage:

    1. Doesn't work well if messages can be arbitrarily large
    2. Requires significant number of file handles

    Additionally, this library makes use of the pickle protocol v5's
    buffer interface in `make_buffer_callback` and `iter_stored_buffers` methods.
    This is to support fast copies of numpy arrays, and other buffers.
    It was chosen over the default pickle serialization method for performance.
    According to the benchmark in `run_benchmark.py`, it
    is around 30x faster for large buffers. For apples to apples comparison, look at how the benchmark performs in
    commit 82e01b395e736e5c1cdaae7e1bd8a7dca3f78435 vs commit 89111ffea55063fd40f1894315b8c26673cbe6ce
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        buf_size: int,
        max_num_elements: int,
        shared_buffer: bool,
        ctx: BaseContext,
    ) -> None:
        self.max_num_elements = max_num_elements
        self.orig_buf_size = buf_size
        # round buffer size up to align size so that every packets buffer starts as aligned
        self.buf_size = buf_size = roundup_to_align(buf_size)
        self._shared_mem_buf = ctx.RawArray(
            ctypes.c_byte, self.buf_size * self.max_num_elements
        )
        self.buf_sizes = ctx.RawArray(ctypes.c_int, self.max_num_elements)
        self.entry_alloc = CyclicAllocator(max_num_elements, ctx)
        self.shared_buffer = shared_buffer

    def put(self, item: Any):
        write_pos = self.entry_alloc.pop_write_pos()
        # now we own the relevant buffers and can write to them
        block_start = write_pos * self.buf_size
        mutable_block_position = Value(block_start)
        item_bytes = cloudpickle.dumps(
            item,
            protocol=cloudpickle.DEFAULT_PROTOCOL,
            buffer_callback=self.make_buffer_callback(mutable_block_position),
        )
        buffer_data_write_size = mutable_block_position.value - block_start
        self.buf_sizes[write_pos] = len(item_bytes)

        # ensure that buffer procol data (starting from beginning of entry)
        # and serialized pickle data (starting from end of entry) won't overlap and clobber each other
        total_write_size = len(item_bytes) + buffer_data_write_size
        if total_write_size > self.orig_buf_size:
            raise ValueError(
                f"Tried to pass item of serialized size {total_write_size}, but PipelineTask.max_message_size is {self.orig_buf_size}"
            )

        pickled_view = memoryview(item_bytes).cast("b")
        mem_view = memoryview(self._shared_mem_buf).cast("b")
        block_end = block_start + self.buf_size
        mem_view[block_end - len(item_bytes) : block_end] = pickled_view
        # makes entry avaliable for reading
        self.entry_alloc.push_read_pos(write_pos)

    def get(self):
        read_pos = self.entry_alloc.pop_read_pos()
        # no producers should be writing to this queue entry due to datastream's semaphores
        num_bytes = int(self.buf_sizes[read_pos])
        read_block_end = (read_pos + 1) * self.buf_size
        mem_view = memoryview(self._shared_mem_buf).cast("b")
        data_bytes = mem_view[read_block_end - num_bytes : read_block_end]

        loaded_data = cloudpickle.loads(
            data_bytes, buffers=self.iter_stored_buffers(read_pos)
        )
        return loaded_data, read_pos

    def free(self, read_pos: int):
        # frees up element for writing
        self.entry_alloc.push_write_pos(read_pos)

    def make_buffer_callback(self, block_position: Value):
        out_of_band_view = memoryview(self._shared_mem_buf).cast("b")
        out_of_band_size_view = memoryview(self._shared_mem_buf).cast("b").cast("i")

        start_pos = block_position.value
        # zero out starting marker
        out_of_band_size_view[start_pos // 4] = END_OF_BUFFER_ID

        def format_buffer(buf_obj):
            mem_view = memoryview(buf_obj)
            if mem_view.shape is None or any(
                dimsize == 0 for dimsize in mem_view.shape  # pylint: disable=E1133
            ):
                # casting not allowed for zero size memory-views, create a new one instead
                src_obj = memoryview(b"").cast("b")
            else:
                src_obj = mem_view.cast("b")
                obj_size = reduce(mul, src_obj.shape, 1)
                src_obj = src_obj.cast("b", (obj_size,))
            src_len = len(src_obj)
            cur_pos = block_position.value
            next_pos = cur_pos + roundup_to_align(ALIGN_SIZE + src_len)
            if next_pos - start_pos > self.buf_size:
                raise ValueError(
                    f"Serialized numpy data coming out to size at least {next_pos - start_pos} in size, but PipelineTask.max_message_size is {self.orig_buf_size}"
                )
            if next_pos - start_pos < self.buf_size:
                # mark current end of buffer by zeroing out size information
                out_of_band_size_view[next_pos // 4] = END_OF_BUFFER_ID

            # set integer size of next chunk
            out_of_band_size_view[cur_pos // 4] = src_len

            # set chunk data
            out_of_band_view[ALIGN_SIZE + cur_pos : ALIGN_SIZE + cur_pos + src_len] = (
                src_obj
            )

            block_position.value = next_pos

        return format_buffer

    def iter_stored_buffers(self, read_pos):
        out_of_band_view = memoryview(self._shared_mem_buf).cast("b")
        out_of_band_size_view = memoryview(self._shared_mem_buf).cast("b").cast("i")
        cur_pos = read_pos * self.buf_size
        end_pos = (read_pos + 1) * self.buf_size
        while cur_pos < end_pos:
            chunk_size = int(out_of_band_size_view[cur_pos // 4])
            if chunk_size == END_OF_BUFFER_ID:
                break
            out_buffer = out_of_band_view[
                ALIGN_SIZE + cur_pos : ALIGN_SIZE + cur_pos + chunk_size
            ]
            if not self.shared_buffer:
                # NOTE: this copies the memory out of shared memory, which is quite expensive
                # but can cause serious problems if references to this shared memory is kept
                # and is also passed to another pipeline step
                out_buffer = bytearray(out_buffer)
            yield out_buffer
            cur_pos += roundup_to_align(ALIGN_SIZE + chunk_size)

    def flush(self, has_error: synchronize.Event):
        # writes already flushed...
        pass


def _put_thread_start(
    _put_thread_sent: tr.Event,
    _put_thread_recived: tr.Event,
    _put_thread_joinable: tr.Event,
    _put_thread_item: List[Any],
    _write_end: mp_connection.Connection,
):
    while True:
        _put_thread_sent.wait()
        _put_thread_sent.clear()
        item = _put_thread_item[0]
        _put_thread_recived.set()

        # write position is reserved, no need to synchronize pipe
        _write_end.send_bytes(item)
        _put_thread_joinable.set()


class AsyncItemPassing:
    """A substitute for mp.Queue
    for one-to-one process sequential communication.
    Unlike mp.Pipe, put() does not block.
    Unlike mp.Queue, this is designed with
    sudden process exiting/erroring in mind.
    """

    def __init__(self, ctx: BaseContext) -> None:
        self._read_end, self._write_end = ctx.Pipe(duplex=False)
        self._put_thread_item = [None]
        # these cannot be pickled for spawn multiprocessing
        self._put_thread_recived = None
        self._put_thread_sent = None
        self._put_thread = None
        self._put_thread_joinable: Optional[synchronize.Event] = None

    def put(self, item):
        if self._put_thread_recived is None:
            self._set_write_end()

        # copies all contents to byte array so that
        # further mutations of the data after this
        # returns does not change the result
        self._put_thread_item[0] = cloudpickle.dumps(item)
        self._put_thread_joinable.clear()
        self._put_thread_sent.set()
        self._put_thread_recived.wait()
        self._put_thread_recived.clear()

    def get(self):
        try:
            return self._read_end.recv()
        except EOFError:
            # file wasn't readable at the moment, but hopefully it will be soon
            pass
        # wait for file to become readable
        select.select([self._read_end.fileno()], [], [])
        return self._read_end.recv()

    def _set_write_end(self):
        self._put_thread_recived = tr.Event()
        self._put_thread_sent = tr.Event()
        self._put_thread_joinable = tr.Event()
        self._put_thread_joinable.set()
        self._put_thread = tr.Thread(
            target=_put_thread_start,
            args=(
                self._put_thread_sent,
                self._put_thread_recived,
                self._put_thread_joinable,
                self._put_thread_item,
                self._write_end,
            ),
            daemon=True,
        )
        self._put_thread.start()

    def flush(self, has_error: synchronize.Event):
        if self._put_thread_joinable is not None:
            while not self._put_thread_joinable.wait(timeout=0.02):
                if has_error.is_set():
                    break


class PipedQueue(AsyncQueue):
    """
    A custom queue implementation based on piped buffers to transfer data.
    """

    def __init__(self, max_num_elements: int, ctx: BaseContext) -> None:
        self.max_num_elements = max_num_elements
        self.queues = [AsyncItemPassing(ctx) for _ in range(max_num_elements)]
        self.entry_alloc = CyclicAllocator(max_num_elements, ctx)

    def put(self, item: Any):
        write_pos = self.entry_alloc.pop_write_pos()
        # this pipe is reserved, will only be read by a single reader
        self.queues[write_pos].put(item)
        # makes entry available for reading
        self.entry_alloc.push_read_pos(write_pos)

    def get(self):
        read_pos = self.entry_alloc.pop_read_pos()
        # only one producer will be writing to this endpoint
        loaded_data = self.queues[read_pos].get()
        return loaded_data, read_pos

    def free(self, read_pos: int):
        # frees up element for writing
        self.entry_alloc.push_write_pos(read_pos)

    def flush(self, has_error: synchronize.Event):
        for queue_ in self.queues:
            queue_.flush(has_error)


class TaskOutput:
    def __init__(
        self,
        *,
        num_upstream_tasks: int,
        packets_in_flight: int,
        error_info: BufferedQueue,
        max_message_size: Optional[int],
        shared_buffer: bool,
        ctx: BaseContext,
    ) -> None:
        # pylint: disable=too-many-arguments
        self.num_tasks_remaining = ctx.Value("i", num_upstream_tasks, lock=True)
        self.last_updated_time = ctx.Value("d", time.monotonic(), lock=False)
        self.queue_len = ctx.Semaphore(value=0)
        self.packets_space = ctx.Semaphore(value=packets_in_flight)
        # using a custom queue implementation rather than multiprocessing.queue
        # because mp.Queue has strange synchronization properties with the semaphores, leading to many bugs
        self.queue = (
            BufferedQueue(max_message_size, packets_in_flight + 1, shared_buffer, ctx)
            if max_message_size is not None
            else PipedQueue(packets_in_flight + 1, ctx)
        )
        self.has_error = ctx.Event()
        self.error_info = error_info

    def iter_results(self) -> Iterable[Any]:
        while True:
            self.queue_len.acquire()  # pylint: disable=consider-using-with
            if self.has_error.is_set():
                raise PropogateErr()

            try:
                item, read_pos = self.queue.get()
            except queue.Empty:
                # only occurs if error occurred or no more producers left
                break

            yield item

            # frees shared memory buffer so that it can be used elsewhere
            self.queue.free(read_pos)

            # this release needs to happen after the yield
            # completes to support full synchronization semantics with packets_in_flight=1
            self.packets_space.release()

            # update last updated time
            self.last_updated_time.value = time.monotonic()

    def put_results(self, iterable: Iterable[Any]):
        iterator = iter(iterable)
        try:
            while True:
                # wait for space to be available on queue before iterating to next item
                # essential for full synchronization semantics with packets_in_flight=1
                self.packets_space.acquire()  # pylint: disable=consider-using-with

                if self.has_error.is_set():
                    raise PropogateErr()

                item = next(iterator)

                self.queue.put(item)
                self.queue_len.release()
        except StopIteration:
            # flush and clean up up queue resources since it is done putting
            self.queue.flush(has_error=self.has_error)
            # normal end of iteration
            with self.num_tasks_remaining.get_lock():
                self.num_tasks_remaining.get_obj().value -= 1
                if self.num_tasks_remaining.get_obj().value == 0:
                    for _i in range(MAX_NUM_WORKERS):
                        self.queue_len.release()

    def set_error(self, task_name, err, traceback_str):
        if not self.has_error.is_set():
            self.has_error.set()
            self.error_info.put((task_name, err, traceback_str))
        # release all consumers and producers semaphores so that they exit quickly
        for _i in range(MAX_NUM_WORKERS):
            self.queue_len.release()
            self.packets_space.release()


def _start_source(
    task: PipelineTask,
    downstream: TaskOutput,
):
    try:
        out_iter = task.generator(**task.constants_dict)
        downstream.put_results(out_iter)
    except Exception as err:  # pylint: disable=broad-except
        tb_str = traceback.format_exc()
        downstream.set_error(task.name, err, tb_str)
        # exiting directly instead of re-raising error, as that would clutter stderr
        # with duplicate tracebacks
        sys.exit(PYTHON_ERR_EXIT_CODE)


def _start_worker(
    task: PipelineTask,
    upstream: TaskOutput,
    downstream: TaskOutput,
):
    try:
        generator_input = upstream.iter_results()
        out_iter = task.generator(generator_input, **task.constants_dict)
        downstream.put_results(out_iter)
    except Exception as err:  # pylint: disable=broad-except
        tb_str = traceback.format_exc()
        # sets upstream and downstream so that error propagates throughout the system
        downstream.set_error(task.name, err, tb_str)
        upstream.set_error(task.name, err, tb_str)
        # exiting directly instead of re-raising error, as that would clutter stderr
        # with duplicate tracebacks
        sys.exit(PYTHON_ERR_EXIT_CODE)


@contextlib.contextmanager
def sighandler(signums: Set[int], processes: List[Union[ForkProcess, SpawnProcess]]):
    def sigterm_handler(signum, _frame):
        # propogate the signal to children processes
        for proc in processes:
            # os.kill just sends a signal like the command line tool
            try:
                os.kill(proc.ident, signum)
            except ProcessLookupError:
                logger.warning(f"Failed to find process {proc.ident}")
        # throw an exception to trigger the exceptional cleanup policy
        raise SignalReceived(signum)

    old_handlings = {}
    for signum in signums:
        old_handlings[signum] = signal.getsignal(signum)
        signal.signal(signum, sigterm_handler)
    did_reset_handling = False
    try:
        yield
    except SignalReceived as sigerr:
        if sigerr.signum in signums:
            # if the signal was raised by our signal handler, then retry the old signal handling method
            # so the end user of the library can handle signals in the way they wish to
            for signum, old_handling in old_handlings.items():
                signal.signal(signum, old_handling)
            did_reset_handling = True
            signal.raise_signal(sigerr.signum)
            # else re-raise error
        else:
            raise sigerr
    finally:
        # only reset handling here if not done in except statement
        if not did_reset_handling:
            for signum, old_handling in old_handlings.items():
                signal.signal(signum, old_handling)


def _start_sink(
    task: PipelineTask,
    upstream: TaskOutput,
):
    try:
        generator_input = upstream.iter_results()
        task.generator(generator_input, **task.constants_dict)
    except Exception as err:  # pylint: disable=broad-except
        tb_str = traceback.format_exc()
        upstream.set_error(task.name, err, tb_str)
        # exiting directly instead of re-raising error, as that would clutter stderr
        # with duplicate tracebacks
        sys.exit(PYTHON_ERR_EXIT_CODE)


def execute_mp(
    tasks: List[PipelineTask],
    spawn_method: SpawnContextName,
    inactivity_timeout: Optional[float] = None,
):
    # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    """
    execute tasks until final task completes.
    Raises error if tasks are inconsistently specified or if
    one of the tasks raises an error.

    Also raises an error if no message passing is observed in any task for
    at least `inactivity_timeout` seconds.
    (useful to kill any stuck jobs in a larger distributed system)
    """
    if not tasks:
        return

    sanity_check_mp_params(tasks)

    if len(tasks) == 1:
        (task,) = tasks
        task.generator(**task.constants_dict)
        return

    ctx = typing.cast(Union[ForkContext, SpawnContext], mp.get_context(spawn_method))

    source_task = tasks[0]
    sink_task = tasks[-1]
    worker_tasks = tasks[1:-1]

    n_total_tasks = sum(task.num_workers for task in tasks)
    # use a BufferedQueue because it synchronizes instantly, unlike PipedQueue or mp.queue
    err_queue = BufferedQueue(ERR_BUF_SIZE, n_total_tasks + 2, False, ctx)
    # number of processes are of the producing task
    data_streams = [
        TaskOutput(
            num_upstream_tasks=t.num_workers,
            packets_in_flight=t.packets_in_flight,
            error_info=err_queue,
            max_message_size=t.max_message_size,
            shared_buffer=t.shared_buffer,
            ctx=ctx,
        )
        for t in tasks[:-1]
    ]
    processes: List[Union[ForkProcess, SpawnProcess]] = [
        ctx.Process(
            target=_start_source,
            args=(source_task, data_streams[0]),
            name=f"{source_task}_{worker_idx}",
        )
        for worker_idx in range(source_task.num_workers)
    ]
    for i, worker_task in enumerate(worker_tasks):
        for worker_idx in range(worker_task.num_workers):
            processes.append(
                ctx.Process(
                    target=_start_worker,
                    args=(worker_task, data_streams[i], data_streams[i + 1]),
                    name=f"{worker_task}_{worker_idx}",
                )
            )

    for worker_idx in range(sink_task.num_workers):
        processes.append(
            ctx.Process(
                target=_start_sink,
                args=(sink_task, data_streams[-1]),
                name=f"{sink_task}_{worker_idx}",
            )
        )

    for process in processes:
        process.start()

    # signal setup must be *after* all new processes are started, so that main processes
    # signal handling won't be copied over to children
    with sighandler({signal.SIGINT, signal.SIGTERM}, processes):
        has_error = False
        try:
            sentinel_map = {proc.sentinel: proc for proc in processes}
            sentinel_set = {proc.sentinel for proc in processes}
            while sentinel_set and not has_error:
                done_sentinels = mp_connection.wait(
                    list(sentinel_set),
                    timeout=(
                        None if inactivity_timeout is None else inactivity_timeout / 10
                    ),
                )
                last_updated_time = max(
                    float(stream.last_updated_time.value) for stream in data_streams
                )
                if inactivity_timeout is not None and not done_sentinels:
                    # this means the timeout ended,
                    # time to check all of the task outputs timers
                    last_updated_time = max(
                        float(stream.last_updated_time.value) for stream in data_streams
                    )
                    if time.monotonic() - last_updated_time > inactivity_timeout:
                        raise InactivityError(
                            f"Last updated time was {time.monotonic() - last_updated_time}s ago, pipeline inactivity timeout is {inactivity_timeout}s."
                        )

                sentinel_set -= set(done_sentinels)
                for done_id in done_sentinels:
                    assert isinstance(
                        done_id, int
                    ), f"mp_connection.wait returned unexpected type: {done_id}"
                    # for some reason needs a join, or the exitcode doesn't sync properly
                    # but it has already exited, so this should finish very quickly
                    sentinel_map[done_id].join()
                    if sentinel_map[done_id].exitcode is None:
                        # unsure what could cause this, but we see it in production sometimes
                        # when an instance is shutting down
                        logger.warning("Child process joined with exitcode None.")
                    elif sentinel_map[done_id].exitcode != 0:
                        # attempts to catch segfaults and other errors that cannot be caught by python (i.g. sigkill)
                        proc_err_msg = f"Process: {sentinel_map[done_id].name} exited with non-zero code {sentinel_map[done_id].exitcode}"
                        for stream in data_streams:
                            stream.set_error(
                                sentinel_map[done_id].name, TaskError(proc_err_msg), ""
                            )
                        has_error = True
                        break

            if has_error:
                # first entry on the error queue should hopefully be the original error, just raise that one single error
                (task_name, task_err, traceback_str), _ = err_queue.get()
                # should only be at most one unique error, just raise it
                # the main error needs to be the main raise for type-based exception catching to work
                raise task_err from TaskError(
                    f"Task; {task_name} errored\n{traceback_str}\n{task_err}"
                )

        except BaseException as err:  # pylint: disable=broad-except
            if not has_error:
                # one of the other processes didn't set an error, so it must be the main process that is the problem
                has_error = True
                tb_str = traceback.format_exc()
                for stream in data_streams:
                    stream.set_error("_main_thread", err, tb_str)

            # joins processes as cleanup if they successfully exited
            # give them a decent amount of time to process their current task and exit cleanly
            for proc in processes:
                proc.join(timeout=15.0)
            # escalate, send sigterm to processes
            for proc in processes:
                proc.terminate()
            # wait for terminate signal to propagate through the processes
            for proc in processes:
                proc.join(timeout=5.0)
            # force kill the processes (only if they are refusing to terminate cleanly)
            for proc in processes:
                proc.kill()
                proc.join()

            raise err



import logging
import threading as tr
import time
import warnings
from collections import deque
from threading import RLock, Semaphore, get_native_id
from typing import Any, Iterable, List, Optional, Set

from .pipeline_task import DEFAULT_BUF_SIZE, InactivityError, PipelineTask, TaskError
from .type_checking import MAX_NUM_WORKERS, sanity_check_mp_params

logger = logging.getLogger(__name__)


class PropogateErr(RuntimeError):
    pass


class TaskOutput:
    def __init__(self, num_upstream_tasks: int, packets_in_flight: int) -> None:
        self.num_tasks_remaining = num_upstream_tasks
        self.queue_len = Semaphore(value=0)
        self.packets_space = Semaphore(value=packets_in_flight)
        self.queue: deque = deque(maxlen=packets_in_flight)
        self.last_updated_time = time.monotonic()
        self.lock = RLock()
        self.error_info = None

    def iter_results(self) -> Iterable[Any]:
        while True:
            self.queue_len.acquire()  # pylint: disable=consider-using-with
            if self.is_errored():
                raise PropogateErr()
            try:
                item = self.queue.popleft()
            except IndexError:
                # only happens when out of results
                return
            yield item

            # this release needs to happen after the yield
            # completes to support full synchronization semantics with packets_in_flight=1
            self.packets_space.release()

            # store the updated time to register that progress was made in the pipeline
            self.last_updated_time = time.monotonic()

    def put_results(self, iterable: Iterable[Any]):
        iterator = iter(iterable)
        try:
            while True:
                # wait for space to be available on queue before iterating to next item
                # essential for full synchronization semantics with packets_in_flight=1
                self.packets_space.acquire()  # pylint: disable=consider-using-with

                if self.is_errored():
                    raise PropogateErr()

                item = next(iterator)

                self.queue.append(item)
                self.queue_len.release()
        except StopIteration:
            # normal end of iteration
            with self.lock:
                self.num_tasks_remaining -= 1
                if self.num_tasks_remaining == 0:
                    for _i in range(MAX_NUM_WORKERS):
                        self.queue_len.release()

    def is_errored(self):
        with self.lock:
            return self.error_info is not None

    def set_error(self, task_name, err):
        with self.lock:
            if self.error_info is None:
                self.error_info = (task_name, err)
        # release all consumers and producers semaphores so that they exit quickly
        for _i in range(MAX_NUM_WORKERS):
            self.queue_len.release()
            self.packets_space.release()


def _start_worker(
    task: PipelineTask,
    upstream: TaskOutput,
    downstream: TaskOutput,
    clean_completed: Set[int],
):
    try:
        constants = {} if task.constants is None else task.constants
        generator_input = upstream.iter_results()
        out_iter = task.generator(generator_input, **constants)
        downstream.put_results(out_iter)
    except BaseException as err:  # pylint: disable=broad-except
        # sets upstream and downstream so that error propagates throughout the system
        downstream.set_error(task.name, err)
        upstream.set_error(task.name, err)
    finally:
        clean_completed.add(get_native_id())


def _start_source(
    task: PipelineTask,
    downstream: TaskOutput,
    clean_completed: Set[int],
):
    try:
        out_iter = task.generator(**task.constants_dict)
        downstream.put_results(out_iter)
    except BaseException as err:  # pylint: disable=broad-except
        downstream.set_error(task.name, err)
    finally:
        clean_completed.add(get_native_id())


def _start_sink(
    task: PipelineTask,
    upstream: TaskOutput,
    clean_completed: Set[int],
):
    try:
        generator_input = upstream.iter_results()
        task.generator(generator_input, **task.constants_dict)
    except BaseException as err:  # pylint: disable=broad-except
        upstream.set_error(task.name, err)
    finally:
        clean_completed.add(get_native_id())


def _warn_parameter_overrides(tasks: List[PipelineTask]):
    for task in tasks:
        if (
            task.max_message_size is not None
            and task.max_message_size != DEFAULT_BUF_SIZE
        ):
            warnings.warn(
                f"Task '{task.name}' overrode default value of max_message_size, and this override is ignored by 'thread' parallelism strategy."
            )


def execute_tr(tasks: List[PipelineTask], inactivity_timeout: Optional[float]):
    # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    """
    execute tasks until final task completes.
    Raises error if tasks are inconsistently specified or if
    one of the tasks raises an error.

    Also raises an error if no message passing is observed in any task for
    at least `inactivity_timeout` seconds.
    (useful to kill any stuck jobs in a larger distributed system)
    """
    if not tasks:
        return

    sanity_check_mp_params(tasks)
    _warn_parameter_overrides(tasks)

    if len(tasks) == 1:
        (task,) = tasks
        task.generator(**task.constants_dict)
        return

    source_task = tasks[0]
    sink_task = tasks[-1]
    worker_tasks = tasks[1:-1]
    clean_completed: Set[int] = set()

    # number of processes are of the producing task
    data_streams = [TaskOutput(t.num_workers, t.packets_in_flight) for t in tasks[:-1]]
    # only one source thread per program
    threads: List[tuple[str, tr.Thread]] = [
        (
            source_task.name,
            tr.Thread(
                target=_start_source,
                args=(source_task, data_streams[0], clean_completed),
            ),
        )
    ]
    for i, worker_task in enumerate(worker_tasks):
        for _ in range(worker_task.num_workers):
            threads.append(
                (
                    worker_task.name,
                    tr.Thread(
                        target=_start_worker,
                        args=(
                            worker_task,
                            data_streams[i],
                            data_streams[i + 1],
                            clean_completed,
                        ),
                    ),
                )
            )

    for _ in range(sink_task.num_workers):
        threads.append(
            (
                sink_task.name,
                tr.Thread(
                    target=_start_sink,
                    args=(sink_task, data_streams[-1], clean_completed),
                ),
            )
        )

    for name, thread in threads:
        thread.start()

    thread_id_to_name = {thread.native_id: name for name, thread in threads}

    sentinel_set = {proc.native_id for _name, proc in threads}
    try:
        while sentinel_set:
            for stream in data_streams:
                # no locking needed to read this value
                # because the other threads are only writing
                last_updated_time = max(
                    float(stream.last_updated_time) for stream in data_streams
                )
                if (
                    inactivity_timeout is not None
                    and time.monotonic() - last_updated_time > inactivity_timeout
                ):
                    raise InactivityError(
                        f"Last updated time was {time.monotonic() - last_updated_time}s ago, pipeline inactivity timeout is {inactivity_timeout}s."
                    )

            done_sentinels = set()
            is_first_thread = True
            for name, thread in threads:
                if thread.native_id not in sentinel_set:
                    continue
                if is_first_thread:
                    timeout = (
                        None if inactivity_timeout is None else inactivity_timeout / 10
                    )
                    is_first_thread = False
                else:
                    timeout = 0
                thread.join(timeout=timeout)
                if not thread.is_alive():
                    done_sentinels.add(thread.native_id)
                    sentinel_set.remove(thread.native_id)

            # check for handled errors
            task_name, err = ("", "")
            for stream in data_streams:
                with stream.lock:
                    if stream.error_info is not None and not isinstance(
                        stream.error_info[1], PropogateErr
                    ):
                        # should only be at most one unique error, just raise it
                        task_name, err = stream.error_info
                        # somehow this error retains the full trackback, no reason to include the thread-specific traceback here
                        raise err

            # check for weird unhandled errors (defensive coding, don't know of real world situations which would cause this)
            for done_id in done_sentinels:
                if done_id not in clean_completed:
                    # attempts to catch various errors that aren't caught by python (i.g. sigkill)
                    proc_err_msg = f"Thead: {done_id} exited improperly"
                    task_name = thread_id_to_name[done_id]
                    for stream in data_streams:
                        stream.set_error(
                            thread_id_to_name[done_id], TaskError(proc_err_msg)
                        )
                    raise TaskError(f"Improper thread exit; {task_name}")

    except BaseException as err:
        # sets errors in streams in case they aren't already set
        for stream in data_streams:
            with stream.lock:
                if stream.error_info is None:
                    stream.set_error("main_task", err)
        # clean up remaining threads so that main process terminates properly
        for _name, thread in threads:
            thread.join(15)
        raise err



from typing import List, Literal, Optional, Tuple, get_args

from pipeline_lib.type_checking import sanity_check_mp_params, type_check_tasks

from .mp_execution import execute_mp
from .pipeline_task import PipelineTask
from .seq_execution import execute_seq
from .tr_execution import execute_tr

ParallelismStrategy = Literal["thread", "process-fork", "process-spawn", "coroutine"]

# list of strings in ParallelismStrategy
PARALLELISM_STRATEGIES: Tuple[str, ...] = get_args(ParallelismStrategy)


def execute(
    tasks: List[PipelineTask],
    parallelism: ParallelismStrategy = "thread",
    type_check_pipeline: bool = True,
    inactivity_timeout: Optional[float] = None,
):
    """
    execute tasks until final task completes.
    Raises exception if
    one of the tasks raises an exception or crashes/segfaults in multiprocessing mode.

    if `type_check_pipeline` is specified, then raises an exception
    if tasks are inconsistently typed or untyped (similar to a harsh mypy analysis).

    Also raises an error if no message passing is observed in any task for
    at least `inactivity_timeout` seconds.
    (useful to kill any stuck jobs in a larger distributed system)
    """

    if not tasks:
        return

    sanity_check_mp_params(tasks)

    if type_check_pipeline:
        type_check_tasks(tasks)

    if parallelism == "thread":
        execute_tr(tasks, inactivity_timeout=inactivity_timeout)
    elif parallelism == "process-spawn":
        execute_mp(tasks, "spawn", inactivity_timeout=inactivity_timeout)
    elif parallelism == "process-fork":
        execute_mp(tasks, "fork", inactivity_timeout=inactivity_timeout)
    elif parallelism == "coroutine":
        assert (
            inactivity_timeout is None
        ), "'coroutine' parallelism does not support inactivity timeout, please choose parallelism='process-fork' or parallelism='process-spawn'"
        execute_seq(tasks)
    else:
        raise ValueError(
            f"`execute`'s parallelism argument must be one of {PARALLELISM_STRATEGIES}"
        )



import multiprocessing as mp
import os
import pickle
from typing import Any, Dict

import numpy as np
import pytest

from pipeline_lib import PipelineTask, execute
from pipeline_lib.execution import ParallelismStrategy

from .example_funcs import *
from .test_utils import all_parallelism_options, sleeper

TEMP_FILENAME = "_test_pipeline_pickle.data"


def save_results(vals: Iterable[int], tmpdir: str) -> None:
    with open(os.path.join(tmpdir, TEMP_FILENAME), "wb") as file:
        pickle.dump(list(vals), file)


def load_results(tempdir: str):
    with open(os.path.join(tempdir, TEMP_FILENAME), "rb") as file:
        return pickle.load(file)


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_execute_basic(parallelism: ParallelismStrategy):
    tasks = [
        PipelineTask(
            generate_numbers,
        ),
        PipelineTask(
            group_numbers,
            constants={"num_groups": 5},
        ),
        PipelineTask(
            sum_numbers,
        ),
        PipelineTask(
            print_numbers,
        ),
    ]
    execute(tasks, parallelism)


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_full_contents_buffering(tmpdir: str, parallelism: ParallelismStrategy):
    """
    Tests the case where the initial work generator fills up the buffer
    and exits, long before the pipeline has completed. Test that the pipeline doesn't
    prematurely terminate the pipeline when the first worker exits.
    """
    tasks = [
        PipelineTask(generate_numbers, packets_in_flight=1000, max_message_size=1000),
        PipelineTask(
            sleeper,
            constants={"sleep_time": 0.1},
            packets_in_flight=1000,
            max_message_size=1000,
            num_workers=2,
        ),
        PipelineTask(save_results, constants=dict(tmpdir=tmpdir)),
    ]
    execute(tasks, parallelism)
    actual_result = sum(load_results(tmpdir))
    expected_result = (101 * 100) // 2
    assert actual_result == expected_result


def add_one_to(vals: Iterable[int], value: mp.Value) -> Iterable[int]:
    for v in vals:
        value.value += 1
        assert value.value == 1
        yield v


def sub_one_to(vals: Iterable[int], value: mp.Value) -> Iterable[int]:
    for v in vals:
        value.value -= 1
        assert value.value == 0
        yield v


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_full_synchronization(tmpdir: str, parallelism: ParallelismStrategy):
    """
    Tests that packets_in_flight=1 forces fully synchronous work
    with no interleaving computation
    """
    val = mp.Value("i", 0, lock=False)
    tasks = [
        PipelineTask(
            generate_numbers,
            packets_in_flight=1,
        ),
        PipelineTask(add_one_to, packets_in_flight=1, constants=dict(value=val)),
        PipelineTask(sub_one_to, packets_in_flight=1, constants=dict(value=val)),
        PipelineTask(add_one_to, packets_in_flight=1, constants=dict(value=val)),
        PipelineTask(sub_one_to, packets_in_flight=1, constants=dict(value=val)),
        PipelineTask(save_results, constants=dict(tmpdir=tmpdir), packets_in_flight=1),
    ]
    execute(tasks, parallelism)
    actual_result = load_results(tmpdir)
    expected_result = [*range(101)]
    assert actual_result == expected_result


def generate_zero_size_np_arrays() -> Iterable[np.ndarray]:
    for _ in range(10):
        val1 = np.zeros((7, 0, 4), dtype="int32")
        yield val1


def consume_nd(inpt: Iterable[np.ndarray]) -> None:
    for _ in inpt:
        pass


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_zero_size_np_arrays(parallelism: bool):
    """zero size buffer passing has some edge cases, so testing that it works
    with a the buffer passing"""
    tasks = [
        PipelineTask(
            generate_zero_size_np_arrays,
            max_message_size=100000,
        ),
        PipelineTask(consume_nd),
    ]
    execute(tasks, parallelism)


def generate_many() -> Iterable[int]:
    yield from range(30000)


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_many_workers_correctness(tmpdir: str, parallelism: ParallelismStrategy):
    """
    Tests that many workers working on lots of data
    eventually returns the correct result, without packet loss or exceptions
    """
    tasks = [
        PipelineTask(
            generate_many,
        ),
        PipelineTask(
            add_const,
            constants={
                "add_val": 5,
            },
            num_workers=15,
            packets_in_flight=15,
        ),
        PipelineTask(
            group_numbers,
            constants={"num_groups": 10},
            num_workers=1,
            packets_in_flight=1,
        ),
        PipelineTask(
            sum_numbers,
            num_workers=16,
            packets_in_flight=20,
        ),
        PipelineTask(save_results, constants=dict(tmpdir=tmpdir)),
    ]
    execute(tasks, parallelism)
    actual_result = sum(load_results(tmpdir))
    expected_result = 450135000
    assert actual_result == expected_result


@pytest.mark.parametrize("parallelism", all_parallelism_options)
def test_many_packets_correctness(tmpdir: str, parallelism: ParallelismStrategy):
    """
    Tests that many workers working on lots of data
    eventually returns the correct result, without packet loss or exceptions
    """
    tasks = [
        PipelineTask(
            generate_many,
            packets_in_flight=10,
        ),
        PipelineTask(
            add_const,
            constants={
                "add_val": 5,
            },
            num_workers=4,
            packets_in_flight=40,
        ),
        PipelineTask(
            group_numbers,
            constants={"num_groups": 10},
            num_workers=4,
            packets_in_flight=10,
        ),
        PipelineTask(
            sum_numbers,
            num_workers=4,
            packets_in_flight=100,
        ),
        PipelineTask(save_results, constants=dict(tmpdir=tmpdir)),
    ]
    execute(tasks, parallelism)
    results = load_results(tmpdir)
    actual_result = sum(results)
    expected_result = 450135000
    assert actual_result == expected_result


N_BIG_MESSAGES = 100
BIG_MESSAGE_SIZE = 200000
BIG_MESSAGE_BYTES = 4 * BIG_MESSAGE_SIZE + 5000


def generate_large_messages() -> Iterable[Dict[str, Any]]:
    for i in range(N_BIG_MESSAGES):
        val1 = np.arange(BIG_MESSAGE_SIZE, dtype="int32").reshape(100, -1) + i
        yield {
            "message_type": "big",
            "message_1_contents": val1,
            "val1_ref": val1,
            "message_2_contents": (np.arange(500, dtype="int64") * i),
        }


def process_message(messages: Iterable[Dict[str, Any]]) -> Iterable[Dict[str, Any]]:
    for msg in messages:
        msg["processed"] = True
        # adds 1 to every element in this and its reference in `val1_ref`
        msg["message_1_contents"] += 1
        yield msg


def sum_arrays(messages: Iterable[Dict[str, Any]]) -> Iterable[int]:
    for msg in messages:
        yield (
            msg["message_1_contents"].astype("int64").sum()
            + msg["val1_ref"].astype("int64").sum()
            + msg["message_2_contents"].astype("int64").sum()
        )


@pytest.mark.parametrize("parallelism", all_parallelism_options)
@pytest.mark.parametrize("n_procs,packets_in_flight", [(1, 1), (1, 4), (4, 16)])
@pytest.mark.parametrize("shared_buffer", [True, False])
def test_many_large_packets_correctness(
    tmpdir: str,
    n_procs: int,
    packets_in_flight: int,
    shared_buffer: bool,
    parallelism: ParallelismStrategy,
):
    tasks = [
        PipelineTask(
            generate_large_messages,
            max_message_size=BIG_MESSAGE_BYTES,
            shared_buffer=shared_buffer,
        ),
        PipelineTask(
            process_message,
            max_message_size=BIG_MESSAGE_BYTES,
            num_workers=n_procs,
            packets_in_flight=packets_in_flight,
            shared_buffer=shared_buffer,
        ),
        PipelineTask(
            process_message,
            # process with piped messages
            num_workers=n_procs,
            packets_in_flight=packets_in_flight,
        ),
        PipelineTask(
            sum_arrays,
            num_workers=n_procs,
            packets_in_flight=packets_in_flight,
        ),
        PipelineTask(save_results, constants=dict(tmpdir=tmpdir)),
    ]
    execute(tasks, parallelism)
    actual_result = sum(load_results(tmpdir))
    expected_result = 4002657512500
    assert actual_result == expected_result


if __name__ == "__main__":
    test_full_synchronization("/tmp", "thread")
